#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Tue Mar  8 15:48:04 2022
# @author: Xiaodong Ming

"""
__init__
===========
To do:
    initialize hydro_raster package
    get sample data

"""
from .Raster import merge, from_tif

def get_sample_data(format_str):
    """ Get sample data for demonstartion
    
    Args:
        format_str: string to specify the data to get
        'tif', 'shp', 

    Returns:
        out_str: string returned as a file name
    """
    import os
    import pkg_resources
    data_path = pkg_resources.resource_filename(__name__, 'sample')
    if format_str == 'tif':
        out_str = os.path.join(data_path, 'CA1_5m.tif')
    elif format_str == 'shp':
        out_str = os.path.join(data_path, 'CA1_overhead_features.shp')
    else:
        raise ValueError('unsupported data format')
    
    return out_str


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Tue Mar 10 15:37:28 2020
# Author: Xiaodong Ming

"""
grid_show
========

To do:

    To visulize grid data, e.g. raster objec(s)

------------------

"""
import os
import copy
import imageio
import numpy as np
import shapefile
import matplotlib.pyplot as plt
#plt.switch_backend('agg')
import matplotlib.colors as colors
from . import spatial_analysis as sp
from matplotlib import cm
from matplotlib.colors import ListedColormap
from matplotlib.patches import Patch
from matplotlib.colors import LightSource
from matplotlib.patches import Polygon
from mpl_toolkits.axes_grid1 import make_axes_locatable
#%% draw inundation map with domain outline
def mapshow(raster_obj=None, array=None, header=None, ax=None, figname=None,
            figsize=None, dpi=300, title=None, cax=True, cax_str=None,
            scale_ratio=1, **kwargs):
    """Display raster data without projection

    Args:
        raster_obj: a Raster object
        array, header: to make Raster object if raster_obj is not given
        figname: the file name to export map, if figname is empty, then the 
            figure will not be saved
        figsize: the size of map
        dpi: The resolution in dots per inch
        vmin and vmax define the data range that the colormap covers
        **kwargs: keywords argument of function imshow

    """
    if raster_obj is not None:
        array = raster_obj.array
        header = raster_obj.header
    # change NODATA_value to nan
    array = array+0
    ind = array == header['NODATA_value']
    if ind.sum()>0:
        array = array.astype('float32')
        array[ind] = np.nan
    # adjust tick label and axis label
    map_extent = sp.header2extent(header)    
    if ax is None:
        fig, ax = plt.subplots(1, figsize=figsize)
    else:
        fig = ax.get_figure()
    img = ax.imshow(array, extent=map_extent, **kwargs)
    
    # add colorbar
	# create an axes on the right side of ax. The width of cax will be 5%
    # of ax and the padding between cax and ax will be fixed at 0.05 inch.
    if cax==True:
        _ = _set_continous_colorbar(ax, img, cax_str)
    ax.axes.grid(linestyle='-.', linewidth=0.2)
    ax.set_aspect('equal', 'box')
    if scale_ratio==1000:
        _adjust_axis_scale(ax)
    if title is not None:
        ax.set_title(title)
    # save figure
    if figname is not None:
        fig.savefig(figname, dpi=dpi, bbox_inches='tight', pad_inches=0)
    # return fig and axis handles
    return fig, ax

def rankshow(raster_obj=None, array=None, header=None, figname=None, 
             figsize=None, dpi=200, ax=None, color='Blues',
             breaks=[0.2, 0.3, 0.5, 1, 2], title=None,
             legend_kw=None, scale_ratio=1, alpha=1, **kwargs):
    """ Display water depth map in ranks defined by breaks

        Args:
            breaks: list of values to define rank. Array values lower than the
                first break value are set as nodata.
            color: color series of the ranks
            ylabelrotation: scalar giving degree to rotate yticklabel
            legend_kw: dict, keyword arguments to set legend. A colobar 
                     rather than a legend be displayed  if legend_kw is None.
            **kwargs: keywords argument of function imshow
        
        Example:
            rankshow(figname=None, figsize=None, 
                     dpi=200, ax=None, color='Blues', 
                     breaks=[0.2, 0.3, 0.5, 1, 2],
                     legend_kw={'loc':'upper left', 'facecolor':None, 
                                'fontsize':'small', 'title':'depth(m)', 
                                'labelspacing':0.1}, 
                     scale_ratio=1,
                     alpha=1)
    """
    if raster_obj is not None:
        array = raster_obj.array
        header = raster_obj.header
    ind = array == header['NODATA_value']
    array = array.astype('float64')+0
    array[ind] = np.nan
    # create color ranks
    array, newcmp, norm= _set_color_rank(array, breaks, color)
    map_extent = sp.header2extent(header)
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = ax.get_figure()
    chm_plot = ax.imshow(array, extent=map_extent, cmap=newcmp, norm=norm, 
                         alpha=alpha, **kwargs)
    if scale_ratio==1000:
        _adjust_axis_scale(ax)
    # create colorbar        
    if legend_kw is not None: # legend
        _set_color_legend(ax, norm, newcmp, legend_kw)
    else:
        _set_rank_colorbar(ax, chm_plot, norm)
    if title is not None:
        ax.set_title(title)
    if figname is not None:
        fig.savefig(figname, dpi=dpi)
    return fig, ax

def hillshade(raster_obj, ax=None, figsize=None, azdeg=315, altdeg=45, vert_exag=1,
              cmap=None, blend_mode='overlay', alpha=1, scale_ratio=1):
    """ Draw a hillshade map
    """
    array = raster_obj.array+0
    if 'NODATA_value' in raster_obj.header:
            array[array == raster_obj.header['NODATA_value']] = np.nan
    array[np.isnan(array)] = np.nanmax(array)
    ls = LightSource(azdeg=azdeg, altdeg=altdeg)
    if cmap is None:
        cmap = plt.cm.gist_earth
    else:
        cmap = plt.get_cmap(cmap)
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig = plt.gcf()
    rgb = ls.shade(array, cmap=cmap, blend_mode=blend_mode, 
                   vert_exag=vert_exag)
    ax.imshow(rgb, extent=raster_obj.extent, alpha=alpha)
    if scale_ratio==1000:
        _adjust_axis_scale(ax)
    return fig, ax

def vectorshow(obj_x, obj_y, figname=None, figsize=None, dpi=300, ax=None,
               scale_ratio=1, **kwargs):
    """plot velocity map of U and V, whose values stored in two raster objects 
    seperately

    """
    X, Y = obj_x.to_points()        
    U = obj_x.array
    V = obj_y.array
    if U.shape!=V.shape:
        raise TypeError('bad argument: the shapes must be the same')
    if 'figsize' in kwargs:
        figsize = kwargs['figsize']
    else:
        figsize = figsize
    if ax is None:
        fig, ax = plt.subplots(1, figsize=figsize)
    else:
        fig = ax.get_figure()
    ax.quiver(X, Y, U, V, **kwargs)
    ax.set_aspect('equal', 'box')
    if scale_ratio==1000:
        _adjust_axis_scale(ax)
    if figname is not None:
        fig.savefig(figname, dpi=dpi)
    return fig, ax

def make_gif(output_file, obj_list=None, header=None, array_3d=None, 
                     time_str=None, breaks=None, fig_names=None,
                     duration=0.5, delete=False, **kwargs):
    """ Create animation of gridded data

    Args:
        mask_header: (dict) header file provide georeference of rainfall mask
        start_date: a datetime object to give the initial date and time of rain
        duration: duration for each frame (seconds)
        cellsize: sclar (meter) the size of rainfall grid cells

    """
    if fig_names is None:
        fig_names = _plot_temp_figs(obj_list, header, array_3d, breaks,
                                time_str, **kwargs)
    images = []
    for fig_name in fig_names:
        images.append(imageio.imread(fig_name))
        if delete:
            os.remove(fig_name)
    # save animation and delete images
    if not output_file.endswith('.gif'):
        output_file = output_file+'.gif'
    imageio.mimsave(output_file, images, duration=duration)

def make_mp4(output_file, obj_list=None, header=None, array_3d=None, 
               time_str=None, breaks=None, fig_names=None, delete=False,
               fps=10, **kwargs):
    """ Create a video file based on a series of grids

    Args:
        obj_list: a list of Raster objects
        header: a header dict providing georeference the grid [not necessary 
                if obj_list was given]
        array_3d: a 3D numpy array storing grid values for each timestep (in 
                 1st dimension), [not necessary if obj_list was given]
        time_str: a list of string to show time information for each frame

    """
    if fig_names is None:
        fig_names = _plot_temp_figs(obj_list, header, array_3d, breaks,
                                    time_str, **kwargs)
    if not output_file.endswith('.mp4'):
        output_file = output_file+'.mp4'
    print(output_file)
    writer = imageio.get_writer(output_file, 'MP4', fps=fps)
    for fig_name in fig_names:
        writer.append_data(imageio.imread(fig_name))
        if delete:
            os.remove(fig_name)
    writer.close()

def plot_shape_file(shp_file, figsize=None, ax=None, color='r', linewidth=0.5,
                    edgecolor='k', facecolor='none', alpha=1, **kw):
    """
    plot a shapefile of polygons or polylines

    Parameters
    ----------
    shp_file : string
        shapefile name.
    figsize : tuple of two int values, optional
        set figsize. The default is None.
    ax : axis object, optional
    color : string or matplotlib color object, The default is 'r'
    linewidth : scalar, optional
        Set linewidth. The default is 0.5.
    edgecolor : string or matplotlib color object, optional
        Set edgecolor of polygon. The default is 'k'.
    facecolor : TYPE, optional
        Set facecolor of polygon. The default is 'none'.
    alpha : scalar value between 0 to 1, optional
        Set transparent rate of polygon. The default is 1.
    **kw : other matplotlib plot arguments

    Returns
    -------
    fig : figure handle
    ax : axis handle

    """
    sf = shapefile.Reader(shp_file)
    if ax is None:
        fig, ax = plt.subplots(1, figsize=figsize)
    else:
        fig = ax.get_figure()
    for one_shape in sf.shapes():
        for i in range(len(one_shape.parts)):
            i_start = one_shape.parts[i]
            if i==len(one_shape.parts)-1:
                i_end = len(one_shape.points)
            else:
                i_end = one_shape.parts[i+1]
            x = [i[0] for i in one_shape.points[i_start:i_end]]
            y = [i[1] for i in one_shape.points[i_start:i_end]]
            if 'POLYGON' in sf.shapeTypeName:
                p = Polygon(one_shape.points, facecolor=facecolor, 
                            edgecolor=edgecolor, alpha=alpha, 
                            linewidth=linewidth, **kw)
                ax.add_patch(p)
            else:
                ax.plot(x, y, color=color, linewidth=linewidth, **kw)
    # ax.set_xlim([sf.bbox[0], sf.bbox[2]])
    # ax.set_ylim([sf.bbox[1], sf.bbox[3]])
    return fig, ax

def _plot_temp_figs(obj_list=None, header=None, array_3d=None,
                    breaks=None, time_str=None, **kwargs):
    """plot a series of temp pictures and save to make animation
    """
    if obj_list is not None:
        header = obj_list[0].header
        array_3d = []
        for grid_obj in obj_list:
            array_3d.append(grid_obj.array)
        array_3d = np.array(array_3d)
    if breaks is None: #plot continous map
        plot_fun = mapshow
    else:
        plot_fun = rankshow
        kwargs['breaks']=breaks
    fig_names = []
    for i in np.arange(array_3d.shape[0]):
        fig_name = 'temp'+str(i)+'.png'
        fig, ax = plot_fun(array=array_3d[i],header=header, **kwargs)
        if type(time_str) is list:
            ax.set_title(time_str[i])
        fig.savefig(fig_name)
        plt.close(fig)
        fig_names.append(fig_name)
    return fig_names
    
#%%
def _set_color_rank(array, breaks, color):
    """Set color rank for array values 
    """
    array[array < breaks[0]] = np.nan
    max_value = np.nanmax(array)
    breaks = copy.deepcopy(breaks)
    if breaks[-1] < max_value:
        breaks.append(max_value)   
    norm = colors.BoundaryNorm(breaks, len(breaks))
    newcolors = cm.get_cmap(color, norm.N)
    newcolors = newcolors(np.linspace(0, 1, norm.N))
    newcmp = ListedColormap(newcolors)     
    return array, newcmp, norm

def _set_rank_colorbar(ax, img, norm):
    """ Set color bar for rankshow on the right of the ax
    """    
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(img, cax=cax)
    y_tick_values = cax.get_yticks()
    boundary_means = [np.mean((y_tick_values[ii],y_tick_values[ii-1])) 
                        for ii in range(1, len(y_tick_values))]
    category_names = [(str(norm.boundaries[ii-1])+'~'+
                       str(norm.boundaries[ii]))
                      for ii in range(1, len(norm.boundaries))]
#    category_names[0] = '<='+str(norm.boundaries[1])
    category_names[-1] = '>'+str(norm.boundaries[-2])
    cax.yaxis.set_ticks(boundary_means)
    cax.yaxis.set_ticklabels(category_names,rotation=0)
    return cax

def _set_color_legend(ax, norm, cmp, legend_kw):
    """ Set color legend attributes

    legend_kw: dict, keyword arguments to set legend, eg: {loc:'lower right', 
                                        bbox_to_anchor:(1,0), facecolor:None}

    """
    category_names = [(str(norm.boundaries[ii-1])+'~'+
                       str(norm.boundaries[ii]))
                      for ii in range(1, len(norm.boundaries))]
#    category_names[0] = '<='+str(norm.boundaries[1])
    category_names[-1] = '>'+str(norm.boundaries[-2])
    ii = 0
    legend_labels = {}
    for category_name in category_names:
        legend_labels[category_name] = cmp.colors[ii,]
        ii = ii+1
    patches = [Patch(color=color, label=label)
               for label, color in legend_labels.items()]
    ax.legend(handles=patches, **legend_kw)
    return ax

def _set_continous_colorbar(ax, img, cax_str=None,
                            loc='right', size='3%', pad=0.05,
                            fontsize='small'):
    """ Set a continous color bar

    cbar = _set_continous_colorbar(ax, img, cax_str=None)

    cax_str: title of color bar

    """
    divider = make_axes_locatable(ax)
    cax = divider.append_axes(loc, size=size, pad=pad)
    cbar = plt.colorbar(img, cax=cax)
    if cax_str is not None:
        cbar.ax.set_xlabel(cax_str, horizontalalignment='left',
                           fontsize=fontsize)
        cbar.ax.xaxis.set_label_coords(0, 1.06)
    return cbar

def _reset_axis_tick(ax, axis_name, tick_space):
    if axis_name == 'x':
        ax_lim = ax.get_xlim()
        
    else:
        ax_lim = ax.get_ylim()
    lim0 = ax_lim[0]-ax_lim[0]%tick_space+tick_space
    new_ticks = np.arange(lim0, ax_lim[1], tick_space)
    if axis_name == 'x':
        ax.set_xticks(new_ticks)
    else:
        ax.set_yticks(new_ticks)

def _adjust_axis_scale(ax, scale_ratio=1000, unit_tag='km'):
    """ Adjust the axis tick to a new new unit 
    """
    xticks = ax.get_xticks()
    dx = xticks[1]-xticks[0]
    if dx < 1000:
        _reset_axis_tick(ax, 'x', 1000)
    yticks = ax.get_yticks()
    dy = yticks[1]-yticks[0]
    if dy < 1000:
        _reset_axis_tick(ax, 'y', 1000)
    xticks = ax.get_xticks()
    yticks = ax.get_yticks()
    xticks_label = (xticks/1000).astype('int64')
    xticks_label_list = [str(x) for x in xticks_label]
    yticks_label = (yticks/1000).astype('int64')
    yticks_label_list = [str(y) for y in yticks_label]
    ax.set_xticklabels(xticks_label_list)
    ax.set_yticklabels(yticks_label_list)
    ax.set_xlabel(unit_tag+' towards east')
    ax.set_ylabel(unit_tag+' towards north')
    return ax


def main():
    print('Package to show grid data')

if __name__=='__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Wed Nov  6 14:33:36 2019

"""

spatial_analysis
================

functions to analyse data in raster and/or feature datasets to replace ArcGridDataProcessing

Assumptions:
    * map unit is meter
    * its cellsize is the same in both x and y direction
    * its reference position is on the lower left corner of the southwest cell

To do:
    * read and write arc

-----------------

"""
__author__ = "Xiaodong Ming"
import os
import gzip
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import rasterio as rio
import shapefile
from rasterio.windows import Window
from matplotlib import cm
from matplotlib.colors import ListedColormap
from matplotlib.patches import Patch
from matplotlib.colors import LightSource
from mpl_toolkits.axes_grid1 import make_axes_locatable
#%%
def arc_header_read(file_name, header_rows=6):
    """ read the header of a asc file as a dictionary

    Args: 
        file_name: (string) file name
        header_rows: (int) number of header rows

    Return:
        dict: header: a dictionary with keys:

                ncols: (int) number of columns

                nrows: (int) number of rows

                xllcorner: (int/float) x-coordinate of the lower left corner of the lower left cell of the grid

                yllcorner: (int/float) y-coordinate of the lower left corner of the bottom left cell of the grid

                cellsize: (int/float) the length of one square cell

                NODATA_value: (int/float)|-9999 the value representing nodata cell

    """
    check_file_existence(file_name)
    # read header
    header = {} # store header information including ncols, nrows,...
    row_ite = 1
    if file_name.endswith('.gz'):
        with gzip.open(file_name, 'rt') as file_h:     
            for line in file_h:
                if row_ite <= header_rows:
                    line = line.split(" ", 1)
                    header_key = line[0]
                    header_value = float(line[1])
                    # convert key to lowercase except nodata_value
                    if len(header_key) < 12:
                        header_key = header_key.lower()
                    header[header_key] = header_value
                else:
                    break
                row_ite = row_ite+1
    else:
        # read header
        with open(file_name, 'rt') as file_h:
            for line in file_h:
                if row_ite <= header_rows:
                    line = line.split(" ", 1)
                    # convert key to lowercase except nodata_value
                    header_key = line[0]
                    header_value = float(line[1])
                    if len(header_key) < 12:
                        header_key = header_key.lower()
                    header[header_key] = header_value
                else:
                    break
                row_ite = row_ite+1
    header['ncols'] = int(header['ncols'])
    header['nrows'] = int(header['nrows'])
    return header

def arcgridread(file_name, header_rows=6, return_nan=True):
    """ Read ArcGrid format raster file

    Args:
        file_name: (str) the file name to read data
        header_rows: (int) the number of head rows of the asc file

    Return:
        numpy array: the data content
        dict: head of the raster to provide reference information of the grid
        tuple: outline extent of the grid (left, right, bottom, top)

    Note: 
        this function can also read compressed gz files

    """
    check_file_existence(file_name)
    header = arc_header_read(file_name, header_rows)
    array = np.loadtxt(file_name, skiprows=header_rows, dtype='float64')
    if return_nan:
        if 'NODATA_value' in header:
            array[array == header['NODATA_value']] = np.nan
    prj_file = file_name[:-4]+'.prj'
    if os.path.isfile(prj_file):
        with open(prj_file, 'r') as file:
            wkt = file.read()
    else:
        wkt = None
    return array, header, wkt

def arcgridwrite(file_name, array, header, compression=False):
    """ write gird data into a ascii file

    Args:
        file_name: (str) the file name to write grid data. A compressed file will automatically add a suffix '.gz'
        array: (int/float numpy array)
        header: (dict) to provide reference information of the grid
        compression: (logic) to inidcate whether compress the ascii file

    Example:

    .. code:: python

        gird = np.zeros((5,10))
        grid[0,:] = -9999
        grid[-1,:] = -9999
        header = {'ncols':10, 'nrows':5, 'xllcorner':0, 'yllcorner':0,
                  'cellsize':2, 'NODATA_value':-9999}
        file_name = 'example_file.asc'
        arcgridwrite(file_name, array, header, compression=False)
        arcgridwrite(file_name, array, header, compression=True)

    """
    array = array+0
    if not isinstance(header, dict):
        raise TypeError('bad argument: header')
    if file_name.endswith('.gz'):
        compression = True
    # create a file (gz or asc)
    if compression:
        if not file_name.endswith('.gz'):
            file_name = file_name+'.gz'
        file_h = gzip.open(file_name, 'wb')
    else:
        file_h = open(file_name, 'wb')
    file_h.write(b"ncols    %d\n" % header['ncols'])
    file_h.write(b"nrows    %d\n" % header['nrows'])
    file_h.write(b"xllcorner    %g\n" % header['xllcorner'])
    file_h.write(b"yllcorner    %g\n" % header['yllcorner'])
    file_h.write(b"cellsize    %g\n" % header['cellsize'])
    file_h.write(b"NODATA_value    %g\n" % header['NODATA_value'])
    array[np.isnan(array)] = header['NODATA_value']
    np.savetxt(file_h, array, fmt='%g', delimiter=' ')
    file_h.close()
    print(file_name + ' created')

def tif_read(file_name, read_extent=None):
    """read tif file (only the 1st band) and return array, header, crs and meta

    Args:
        file_name (str): tif file name
        read_extent (dict, optional): extent window (left, right, bottom, top 
            in map coordinates) to read part of the raster. Defaults to None.

    Returns:
        _type_: _description_
    """
    with rio.open(file_name) as src:
        ras_meta = src.meta
    if read_extent is None:
        with rio.open(file_name) as src:
            masked_array = src.read(1, masked=True)
    else:
        X = np.array(read_extent[:2])
        Y = np.array(read_extent[2:])
        geo_transform = ras_meta['transform']
        rows, cols = map2sub(X, Y, geo_transform)
        slice_r = slice(rows.min(), rows.max())
        slice_c = slice(cols.min(), cols.max())
        window = Window.from_slices(slice_r, slice_c)
        with rio.open(file_name) as src:
            masked_array = src.read(1, masked=True, 
                                    window=window)
            win_transform = src.window_transform(window)
        ras_meta['height'] = masked_array.data.shape[0]
        ras_meta['width'] = masked_array.data.shape[1]
        ras_meta['transform'] = win_transform 
    array = masked_array.data+0.0
    array[array == masked_array.fill_value] = np.nan
    header = meta2header(ras_meta)   
    crs = ras_meta['crs']
    return array, header, crs, ras_meta

def byte_file_read(file_name):
    """ Read file from a bytes object
    """
    # read header
    header = {} # store header information including ncols, nrows, ...
    num_header_rows = 6
    for _ in range(num_header_rows):
        line = file_name.readline()
        line = line.strip().decode("utf-8").split(" ", 1)
        header[line[0]] = float(line[1])
        # read value array
    array  = np.loadtxt(file_name, skiprows=num_header_rows, 
                        dtype='float64')
    array[array == header['NODATA_value']] = float('nan')
    header['ncols'] = int(header['ncols'])
    header['nrows'] = int(header['nrows'])
    return array, header

def read_shapefile_as_list(shp_name):
    # read shapefile as a list of dict using pyshp
    # return a list of dict giving geometry and a list of dict giving fields
    sf = shapefile.Reader(shp_name)
    shapes_all = sf.shapes()
    shape_dict_list = []
    for one_shape in shapes_all:
        shape_dict = {'type':one_shape.shapeTypeName, 
                      'coordinates':one_shape.points}
        shape_dict_list.append(shape_dict)
    records_all = sf.records()
    record_dict_list = [x.as_dict() for x in records_all]
    return shape_dict_list, record_dict_list

#%% Combine raster files
def combine_raster(asc_files, num_header_rows=6):
    """Combine a list of asc files to a DEM Raster

    Args:   
        asc_files: a list of asc file names. All raster files have the same 
        cellsize.
    """
    # default values for the combined Raster file
    xllcorner_all = []
    yllcorner_all = []
    extent_all =[]
    # read header
    for file in asc_files:
        header0 = arc_header_read(file, num_header_rows)
        extent0 = header2extent(header0)
        xllcorner_all.append(header0['xllcorner'])
        yllcorner_all.append(header0['yllcorner'])
        extent_all.append(extent0)
    cellsize = header0['cellsize']
    if 'NODATA_value' in header0.keys():
        NODATA_value = header0['NODATA_value']
    else:
        NODATA_value = -9999
    xllcorner_all = np.array(xllcorner_all)
    xllcorner = xllcorner_all.min()
    yllcorner_all = np.array(yllcorner_all)
    yllcorner = yllcorner_all.min()
    extent_all = np.array(extent_all)
    x_min = np.min(extent_all[:,0])
    x_max = np.max(extent_all[:,1])
    y_min = np.min(extent_all[:,2])
    y_max = np.max(extent_all[:,3])
#    extent = (x_min, x_max, y_min, y_max)
#    print(extent)
    nrows = int((y_max-y_min)/cellsize)
    ncols = int((x_max-x_min)/cellsize)
    header = header0.copy()
    header['xllcorner'] = xllcorner
    header['yllcorner'] = yllcorner
    header['ncols'] = ncols
    header['nrows'] = nrows
    header['NODATA_value'] = NODATA_value
    array = np.zeros((nrows ,ncols))+NODATA_value
    print(array.shape)
    for file in asc_files:
        array0, header0, _ = arcgridread(file, num_header_rows)
        extent0 = header2extent(header0)
        x0 = extent0[0]+header0['cellsize']/2
        y0 = extent0[3]-header0['cellsize']/2
        row0, col0 = map2sub(x0, y0, header)
        array[row0:row0+header0['nrows'],
              col0:col0+header0['ncols']] = array0
    array[array == header['NODATA_value']] = float('nan')
    extent = header2extent(header)
    return array, header, extent

#%% ------------------------Supporting functions-------------------------------
def check_file_existence(file_name):
    """ check whether a file exists
    """
    try:
        file_h = open(file_name, 'r')
        file_h.close()
    except FileNotFoundError:
        raise

def header2extent(header):
    """ convert a header dict to a 4-element tuple (left, right, bottom, top) 
    all four elements shows the coordinates at the edge of a cell, not center
    """
    left = header['xllcorner']
    right = header['xllcorner']+header['ncols']*header['cellsize']
    bottom = header['yllcorner']
    top = header['yllcorner']+header['nrows']*header['cellsize']
    extent = (left, right, bottom, top)
    return extent

def meta2header(ras_meta):
    """ Transfer rasterio meta object to a header dict
    """
    ncols = ras_meta['width']
    nrows = ras_meta['height']
    transform = ras_meta['transform']
    cellsize = transform[0]
    xllcorner = transform[2]
    yllcorner = transform[5] - cellsize*nrows
    header = {}
    header['ncols'] = ncols
    header['nrows'] = nrows
    header['xllcorner'] = xllcorner
    header['yllcorner'] = yllcorner
    header['cellsize'] = cellsize
    header['NODATA_value'] = ras_meta['nodata']
    return header

def shape_extent_to_header(shape, extent, nan_value=-9999):
    """ Create a header dict with shape and extent of an array
    """
    ncols = shape[1]
    nrows = shape[0]
    xllcorner = extent[0]
    yllcorner = extent[2]
    cellsize_x = (extent[1]-extent[0])/ncols
    cellsize_y = (extent[3]-extent[2])/nrows
    if cellsize_x != cellsize_y:
        raise ValueError('extent produces different cellsize in x and y')
    cellsize = cellsize_x
    header = {'ncols':ncols, 'nrows':nrows,
              'xllcorner':xllcorner, 'yllcorner':yllcorner,
              'cellsize':cellsize, 'NODATA_value':nan_value}
    return header

def map2sub(X, Y, geo_ref):
    """convert map coordinates to subscripts of the raster 
    array according to geo_ref which is either a header or
    a geo_transform object from raster

    Args:
        X (scalar or numpy.ndarray): Input X coordinate
        Y (scalar or numpy.ndarray): Input Y coordinate
        geo_ref (_type_): _description_

    Returns:
        tuple (int, int): rows and cols in the array
    """
    X = np.array(X)
    Y = np.array(Y)
    if type(geo_ref) is dict: #header
        cellsize = geo_ref['cellsize']
        x0 = geo_ref['xllcorner']+0.5*cellsize
        y0 = geo_ref['yllcorner']+(geo_ref['nrows']-0.5)*cellsize
        rows = (y0-Y)/cellsize # row and col number starts from 0
        cols = (X-x0)/cellsize
    else: # geo_transform
        cellsize = geo_ref[0]
        x_ulcorner = geo_ref[2]
        y_ulcorner = geo_ref[5]
        cols = (X-x_ulcorner)/cellsize-0.5
        rows = (y_ulcorner-Y)/cellsize-0.5

    if isinstance(rows, np.ndarray):
        rows = np.round(rows).astype('int64')
        cols = np.round(cols).astype('int64')
    else:
        rows = int(rows)
        cols = int(cols)
    return rows, cols

def sub2map(rows, cols, header):
    """convert subscripts of a matrix to map coordinates rows,
    cols: subscripts of the data matrix, starting from 0

    Args: 
        rows: rows in the array
        cols: cols in the array

    Returns: 
        X and Y coordinate values

    """
    #x and y coordinate of the centre of the first cell in the matrix
    if not isinstance(rows, np.ndarray):
        rows = np.array(rows)
        cols = np.array(cols)        
    extent = header2extent(header)
    left = extent[0] #(left, right, bottom, top)
    top = extent[3]
    X = left + (cols+0.5)*header['cellsize']
    Y = top  - (rows+0.5)*header['cellsize']  
    return X, Y

#% Extent compare between two Raster objects
def compare_extent(extent0, extent1):
    """Compare and show the difference between two Raster extents

    Args:
        extent0, extent1: objects or extent dicts to be compared
        displaye: whether to show the extent in figures

    Return:
        int: 0 extent0>=extent1; 1 extent0<extent1; 2 extent0 and extent1 have
        intersections

    """
    logic_left = extent0[0]<=extent1[0]
    logic_right = extent0[1]>=extent1[1]
    logic_bottom = extent0[2]<=extent1[2]
    logic_top = extent0[3]>=extent1[3]
    logic_all = logic_left+logic_right+logic_bottom+logic_top
    if logic_all == 4:
        output = 0
    elif logic_all == 0:
        output = 1
    else:
        output = 2
        print(extent0)
        print(extent1)
    return output
   
def extent2shape_points(extent):
    """Convert extent to a two-col numpy array of shape points
    """
    #extent = (left, right, bottom, top)
    shape_points = np.array([[extent[0], extent[2]], 
                             [extent[1], extent[2]], 
                             [extent[1], extent[3]], 
                             [extent[0], extent[3]]])
    return shape_points

def _adjust_map_extent(extent, relocate=True, scale_ratio=1):
    """
    Adjust the extent (left, right, bottom, top) to a new staring point and 
    new unit. extent values will be divided by the scale_ratio

    Example:
        if scale_ratio = 1000, and the original extent unit is meter, then the 
        unit is converted to km, and the extent is divided by 1000
    """
    if relocate:
        left = 0 
        right = (extent[1]-extent[0])/scale_ratio
        bottom = 0
        top = (extent[3]-extent[2])/scale_ratio
    else:
        left = extent[0]/scale_ratio
        right = extent[1]/scale_ratio
        bottom = extent[2]/scale_ratio
        top = extent[3]/scale_ratio
    return (left, right, bottom, top)

def _set_colorbar(ax,img,norm):
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(img, cax=cax)
    y_tick_values = cax.get_yticks()
    boundary_means = [np.mean((y_tick_values[ii],y_tick_values[ii-1])) 
                        for ii in range(1, len(y_tick_values))]
    category_names = [(str(norm.boundaries[ii-1])+'~'+
                       str(norm.boundaries[ii]))
                      for ii in range(1, len(norm.boundaries))]
    category_names[0] = '<='+str(norm.boundaries[1])
    category_names[-1] = '>'+str(norm.boundaries[-2])
    cax.yaxis.set_ticks(boundary_means)
    cax.yaxis.set_ticklabels(category_names,rotation=0)
    return cax

def _set_color_legend(ax, norm, cmp,
                      loc='lower right', bbox_to_anchor=(1,0),
                      facecolor=None):
    category_names = [(str(norm.boundaries[ii-1])+'~'+
                       str(norm.boundaries[ii]))
                      for ii in range(1, len(norm.boundaries))]
    category_names[0] = '<='+str(norm.boundaries[1])
    category_names[-1] = '>'+str(norm.boundaries[-2])
    ii = 0
    legend_labels = {}
    for category_name in category_names:
        legend_labels[category_name] = cmp.colors[ii,]
        ii = ii+1
    patches = [Patch(color=color, label=label)
               for label, color in legend_labels.items()]
    ax.legend(handles=patches, loc=loc,
              bbox_to_anchor=bbox_to_anchor,
              facecolor=facecolor)
    return ax

def _insensitive_header_keys(header_dict):
    """
    

    Parameters
    ----------
    header_dict : change the string value 
        DESCRIPTION.

    Returns
    -------
    None.

    """

def main():
    print('Fucntions to process asc data')

if __name__=='__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on Wed Feb  9 16:47:46 2022
# @author: Xiaodong Ming

"""
setup the package
To do:
    setup hydro-raster package
Created on Wed Apr  1 15:03:58 2020

@author: Xiaodong Ming
"""
from setuptools import setup, find_packages
from os import path
here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
      name='hydro_raster',
      version='0.0.10',
      description='To process raster data for hydrological/hydrodynamic modelling',
      url='https://github.com/mingxiaodong/hydro-raster',
      author='Xiaodong Ming',
      author_email='x.ming@lboro.ac.uk',
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=['Development Status :: 3 - Alpha',
                   'Programming Language :: Python :: 3',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 3',
                   'Operating System :: OS Independent',],
      keywords='Raster Hydrological Hydrodynamic',
      license='LICENSE.txt',
      packages=find_packages(),
      include_package_data=True,
      package_data={'hydro_raster': ['sample/CA1_5m.tif',
                                     'sample/CA1_overhead_features.shp',]},
      install_requires=['rasterio', 'scipy', 'shapely', 'pyshp',
                        'matplotlib', 'numpy', 'pandas', 'imageio', 'geopandas'],
      python_requires='>=3.6')

"""
Required python package

#Sometimes you’ll want to use packages that are properly arranged with 
#setuptools, but aren’t published to PyPI. In those cases, you can specify a 
#list of one or more dependency_links URLs where the package can be downloaded,
#along with some additional hints, and setuptools will find and install the
#package correctly.
#For example, if a library is published on GitHub, you can specify it like:
setup(
    ...
    dependency_links=['http://github.com/user/repo/tarball/master#egg=package-1.0']
    ...
)
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Created on Tue Mar 31 16:20:13 2020
# Author: Xiaodong Ming

"""
Raster
=======

To do:
    1. Read, write, analyze, and visualise gridded raster data
    2. Convert shape data to raster data
    3. Convert raster data to shapefile
---------------

"""
import os
import copy
import math
import shapely
import numpy as np
import rasterio as rio
import geopandas as gpd

from rasterio.features import shapes as rio_shapes
from shapely.geometry import shape as shapely_shape
from scipy import interpolate
from . import spatial_analysis as sp
from . import grid_show as gs
#%% *******************************To deal with raster data********************
#   ***************************************************************************    
class Raster(object):
    """    
    To deal with raster data with a ESRI ASCII or GTiff format

    Attributes:
        source_file: file name to read grid data 
        output_file: file name to write a raster object 
        array: a numpy array storing grid cell values 
        header: a dict storing reference information of the grid, with keys:
            nrows, ncols, xllcorner, yllcorner, cellsize, NODATA_value
        extent: a tuple storing outline limits of the raster (left, right, 
            bottom, top)
        shape: shape of the Raster value array
        cellsize: the length of each square cell
        extent_dict: a dictionary storing outline limits of the raster  
        wkt: (string) the Well-Known_Text (wkt) projection information  
        
    """
#%%======================== initialization function ===========================   
    def __init__(self, source_file=None, array=None, header=None, 
                 xllcorner=0, yllcorner=0, cellsize=100, NODATA_value=-9999,
                 crs=None, num_header_rows=6, read_extent=None):
        """Initialise the object

        Args:
            source_file: name of a asc/tif file if a file read is needed
            array: values in each raster cell [a numpy array]
            header: georeference of the raster [a dictionary containing 6 keys]
                nrows, nclos [int]
                cellsize, xllcorner, yllcorner
                NODATA_value
            crs: coordinate reference system, either epsg(epsg code, int) or
                wkt(string), or a rasterio.crs object
            read_extent (dict, optional): extent window (left, right, bottom, top 
                in map coordinates) to read part of the tif. Defaults to None.
        Example: define a raster object with a random array
            array = np.random.rand(10, 10)
            header = {'ncols':array.shape[1], 'nrows':array.shape[0],
                      'xllcorner':0, 'yllcorner':1,
                      'cellsize':100, 'NODATA_value':-9999}
            obj_ras = Raster(array=array, header=header)
            obj_ras.mapshow() %plot map

        """
        # get data from file or arguments
        if type(source_file) is str: # tif, txt, asc, or gz file
            if source_file.endswith('.tif'): # only read the first band
                array, header, crs, ras_meta = sp.tif_read(source_file, read_extent)
                self.meta = ras_meta
            else:
                array, header, crs = sp.arcgridread(source_file,
                                                    num_header_rows)                
            self.source_file = source_file
        elif type(source_file) is bytes:  # try a binary file-like object
            array, header = sp.byte_file_read(source_file)
            self.source_file = None
        
        # projection
        if crs is not None:
            self.set_crs(crs)
        # header and extent
        if header is None:
            header = {'ncols':array.shape[1], 'nrows':array.shape[0],
                      'xllcorner':xllcorner, 'yllcorner':yllcorner,
                      'cellsize':cellsize, 'NODATA_value':NODATA_value}
        self.header = header
        self.extent = sp.header2extent(header)
        self.extent_dict = {'left':self.extent[0], 'right':self.extent[1],
                            'bottom':self.extent[2], 'top':self.extent[3]}
        
        # raster value array
        ind = array == header['NODATA_value']
        if ind.sum() > 0:
            self.array = array+0.0
            self.array[ind] = np.nan
        else:
            self.array = array
        # shape and cellsize
        self.shape = self.array.shape
        if array.shape != (header['nrows'], header['ncols']):
            raise ValueError(('shape of array is not consistent with '
                              'nrows and ncols in header'))
        self.cellsize = header['cellsize']
        self.NODATA_value = header['NODATA_value']
        # meta
        if not hasattr(self, 'meta'):
            self.set_meta()
        # num_valid_cells
        self.num_valid_cells = np.sum(~np.isnan(self.array))
        self.size = self.array.size
        
        # summary
        self.get_summary()
        
        self.to_rasterio = self.write_tif
        
        self.copy = self.duplicate

            
#%%============================= Spatial analyst ==============================   
    def get_summary(self):
        """Get information summary of the object

        Returns
        -------
        summary : dict
            information summary of the object.

        """
        summary = copy.deepcopy(self.header)
        summary['num_valid_cells'] = self.num_valid_cells
        if hasattr(self, 'crs'):
            summary['crs'] = self.crs.data
        if hasattr(self, 'source_file'):
            summary['source_file'] = self.source_file
        self._summary = summary
        return summary

    def to_int(self, dtype='int32'):
        """Convert data array to int type
        nan value will be replaced as NODATA_value, default: -9999 
        

        Returns
        -------
        None.

        """
        new_obj = self.copy()
        new_obj.array[np.isnan(self.array)] = self.NODATA_value
        new_obj.array = np.around(new_obj.array).astype(dtype)
        return new_obj

    def set_crs(self, crs):
        """set reference coordinate system
        Parameters
        ----------
        crs : int|string|rasterio.crs object
            epsg code|wkt|asterio.crs object
        """
        if type(crs) is str:
            self.crs = rio.crs.CRS.from_wkt(crs)
        elif type(crs) is int:
            self.crs = rio.crs.CRS.from_epsg(crs)
        elif type(crs) is rio.crs.CRS:
            self.crs = crs
        else:
            raise IOError('crs must be int|string|rasterio.crs object')

    def rect_clip(self, clip_extent, return_slice=False):
        """clip raster according to a rectangle extent

        Args:
            clip_extent: list of [left, right, bottom, top]
            return_slice: logical. If True, the slices of object to clip array
              will be returned as an additional return value.

        Return:
            Raster: a new raster object
        """
        new_extent = list(clip_extent)
        if clip_extent[0] < self.extent[0]:
            new_extent[0] = self.extent[0]
        if clip_extent[1] > self.extent[1]:
            new_extent[1] = self.extent[1]
        if clip_extent[2] < self.extent[2]:
            new_extent[2] = self.extent[2]
        if clip_extent[3] > self.extent[3]:
            new_extent[3] = self.extent[3]
        new_extent = tuple(new_extent)
        X = np.array(new_extent[0:2])
        Y = np.array(new_extent[2:4])
        cellsize = self.cellsize
        header_new = copy.deepcopy(self.header)
        X_centre = np.array([X.min()+cellsize/2, X.max()-cellsize/2])
        Y_centre = np.array([Y.min()+cellsize/2, Y.max()-cellsize/2])
        rows, cols = sp.map2sub(X_centre, Y_centre, self.header)
        # get x and y centre coords in the original raster
        x_centre, y_centre = sp.sub2map(rows, cols, self.header)
        xllcorner = min(x_centre)-cellsize/2
        yllcorner = min(y_centre)-cellsize/2
        max_row = max(rows)+1
        max_col = max(cols)+1
        min_row = min(rows) 
        min_col = min(cols)
        loc = (slice(min_row, max_row), slice(min_col, max_col))
        array_new = self.array[loc]
        header_new['nrows'] = array_new.shape[0]
        header_new['ncols'] = array_new.shape[1]
        header_new['xllcorner'] = xllcorner
        header_new['yllcorner'] = yllcorner
        obj_new = Raster(array=array_new, header=header_new)
        if hasattr(self, 'crs'):
            obj_new.set_crs(self.crs)
        if return_slice:
            return obj_new, loc
        else:
            return obj_new
    
    def clip(self, clip_mask=None):
        """clip raster according to a mask

        Args:
            mask: 1. string name of a shapefile or 2. 2-col numpy array giving
                X and Y coords in each column to shape the mask polygon
        
        Return:
            Raster: a new raster object
        """
        from rasterio import mask
        ds_rio = self.to_rasterio_ds()
        if type(clip_mask) is str:
            try:
                shape_data = gpd.read_file(clip_mask)
                shape_geoms = list(shape_data.geometry)
            except:
                ds_rio.close()
                raise IOError(f'cannot find {clip_mask}')
        elif type(clip_mask) is np.ndarray:
            shape_geoms = {'type':'Polygon', 'coordinates':[clip_mask]}
            shape_geoms = [shape_geoms]
        else:
            ds_rio.close()
            raise IOError('mask must be either a string or a numpy array')
        
        out_image, out_transform = mask.mask(ds_rio, shape_geoms, crop=True) #
        ds_rio.close()
        array_new = out_image[0]
        cellsize = out_transform[0]
        xllcorner = out_transform[2]
        yllcorner = out_transform[5]- cellsize*array_new.shape[0]
        header_new = copy.deepcopy(self.header)
        header_new['nrows'] = array_new.shape[0]
        header_new['ncols'] = array_new.shape[1]
        header_new['xllcorner'] = xllcorner
        header_new['yllcorner'] = yllcorner
        obj_new = Raster(array=array_new, header=header_new)
        if hasattr(self, 'crs'):
            obj_new.crs = self.crs
        return obj_new
    
    def rasterize(self, shp_filename=None, shape_data=None, attribute=None, 
                  include_nan=False):
        """
        rasterize a shapefile to the raster object and return a bool array
            with Ture value in and on the polygon/polyline or return an array 
            with values burned from one shapefile attritute whose name
            is given by attr_name

        Parameters
        ----------
        shp_filename : str|shapefile name. Must has the same 
            crs with the raster file
        shape_data: geopandas dataframe or shapely geometry. if shp_filename 
            not given, this input will be used as the shape data
        attribute : str or a list of digit values to rasterize with shape, 
            name of the shape attribute to be burned into the raster. 
            The default is None.
        include_nan: logical, optional
            whether include index of nan cells.
            The default is False.

        Returns
        -------
        out_array : numpy array
            Array providing values of one attribute of the shapefile

        """
        from rasterio import features
        if shp_filename is None:
            if isinstance(shape_data, gpd.geodataframe.GeoDataFrame):
                geoms = list(shape_data.geometry)
            elif isinstance(shape_data, shapely.geometry.base.BaseGeometry):
                geoms = [shape_data]
            else:
                raise IOError('shape_data must be either a geopandas dataframe'
                              ' or a shapely object')
        else:
            shape_data = gpd.read_file(shp_filename)
            geoms = list(shape_data.geometry)
        if type(attribute) is str:
            attr_list = list(shape_data[attribute])
        elif type(attribute) is list:
            attr_list = attribute
        else:
            attr_list = None
    
        if attribute is None:
            shapes_iterable = [(geom, i) for i, geom in enumerate(geoms)]
        else:
            if attr_list is None:
                attr_list = list(attribute)
            shapes_iterable = [(geom, val) for geom, val in zip(geoms, attr_list)]
        
        out_arr = np.full(self.shape, np.nan)
        geo_transform = self.meta['transform']      
        burned = features.rasterize(shapes=shapes_iterable, fill=0, 
                                    out=out_arr, transform=geo_transform)
        if include_nan:
            burned[np.isnan(burned)] = 1
        burned[burned == self.meta['nodata']] = np.nan
        out_array = burned
        return out_array
    
    def line2sub(self, line_coords):
        nrows, ncols = self.shape
        line_coords = np.array(line_coords)
        r, c = sp.map2sub(line_coords[:, 0], line_coords[:, 1], self.header)
        subs_all = np.c_[r, c]
        line_subs = []
        for N in np.arange(r.size-1):
            subs_p0 = subs_all[N]
            subs_p1 = subs_all[N+1]
            sub_range = np.abs(subs_p1-subs_p0)
            sub_steps = sub_range.max()+1
            seg_rows = np.linspace(subs_p0[0], subs_p1[0], sub_steps)
            seg_rows = seg_rows.round().astype('int32')
            seg_cols = np.linspace(subs_p0[1], subs_p1[1], sub_steps)
            seg_cols = seg_cols.round().astype('int32')
            seg_subs = np.c_[seg_rows, seg_cols]
            if (seg_subs[-1]==subs_p1).all():
                line_subs.append(seg_subs[:-1])
            else:
                line_subs.append(seg_subs)
        line_subs = np.concatenate(line_subs)
        sub_rows = line_subs[:, 0]
        sub_cols = line_subs[:, 1]
        ind_valid_r = np.logical_and(sub_rows>=0, sub_rows<nrows)
        ind_valid_c = np.logical_and(sub_cols>=0, sub_cols<ncols)
        ind_valid = np.logical_and(ind_valid_r, ind_valid_c)
        return sub_rows[ind_valid], sub_cols[ind_valid]
        
        
    def resample(self, new_cellsize, method='bilinear'):
        """ Resample the raster object to a new resolution
        
        Args:
            new_cellsize: scalar, the resoltion of the new raster object
            method: string, one of the values including 'nearest', 'bilinear',
            'cubic', 'cubic_spline', 'lanczos', 'average', 'mode', 'gauss',
            'max', 'min', 'med', 'q1', 'q3' 
        
        Return:
            Raster object
        """
        from rasterio.enums import Resampling
        method_list = ['nearest', 'bilinear', 'cubic', 'cubic_spline',
                       'lanczos', 'average', 'mode', 'gauss', 'max', 'min',
                       'med', 'q1', 'q3']
        ind = method_list.index(method)
        upscale_factor = self.cellsize/new_cellsize
        ds_rio = self.to_rasterio_ds()
        new_shape = (1, np.around(ds_rio.height * upscale_factor).astype('int'),
                        np.around(ds_rio.width * upscale_factor).astype('int'))
        resampling_method = Resampling(ind)
        data = ds_rio.read(out_shape=new_shape, resampling=resampling_method)
        data = data[0]
        # scale image transform
        transform = ds_rio.transform * ds_rio.transform.scale(
                    (ds_rio.width / data.shape[-1]),
                    (ds_rio.height / data.shape[-2]))
        ds_rio.close()
        new_header = copy.deepcopy(self.header)
        new_header['cellsize'] = new_cellsize
        new_header['nrows'] = data.shape[0]
        new_header['ncols'] = data.shape[1]
        new_header['xllcorner'] = transform[2]
        new_header['yllcorner'] = transform[5]-data.shape[0]*new_cellsize
        obj_new = Raster(array=data, header=new_header)
        if hasattr(self, 'crs'):
            obj_new.crs = self.crs
        return obj_new
            
    def point_interpolate(self, points, values, method='nearest'):
        """ Interpolate values of 2D points to all cells on the Raster object

        2D interpolate

        Args:
            points: ndarray of floats, shape (n, 2)
                Data point coordinates. Can either be an array of shape (n, 2), 
                or a tuple of ndim arrays.
            values: ndarray of float or complex, shape (n, )
                Data values.
            method: {'linear', 'nearest', 'cubic'}, optional
                Method of interpolation.

        """
        grid_x, grid_y = self.to_points()
        array_interp = interpolate.griddata(points, values, (grid_x, grid_y),
                                            method=method)
        obj_new = copy.deepcopy(self)
        obj_new.array = array_interp
        return obj_new
    
    def grid_interpolate(self, value_grid, method='nearest'):
        """ Interpolate values of a grid to all cells on the Raster object

        2D interpolate

        Args:
            value_grid: a grid file string or Raster object 
            method: {'linear', 'nearest', 'cubic'}, optional
                Method of interpolation.

        Return: 
            numpy array: the interpolated grid with the same size of the self 
                object
        """
        if type(value_grid) is str:
            value_grid = Raster(value_grid)
        points_x, points_y = value_grid.to_points()
        points = np.c_[points_x.flatten(), points_y.flatten()]
        values = value_grid.array.flatten()
        ind_nan = ~np.isnan(values)
        grid_x, grid_y = self.to_points()
        array_interp = interpolate.griddata(points[ind_nan, :],
                                            values[ind_nan], (grid_x, grid_y),
                                            method=method)
        return array_interp
    
    def grid_resample_nearest(self, newsize):
        """resample a grid to a new grid resolution via nearest interpolation

        Alias: GridResample
        """
        if isinstance(newsize, dict):
            header = newsize.copy()
        else:
            oldsize = self.header['cellsize']
            header = copy.deepcopy(self.header)
            header['cellsize'] = newsize
            ncols = math.floor(oldsize*self.header['ncols']/newsize)
            nrows = math.floor(oldsize*self.header['nrows']/newsize)
            header['ncols'] = ncols
            header['nrows'] = nrows
        #centre of the first cell in array
        x11 = header['xllcorner']+0.5*header['cellsize']
        y11 = header['yllcorner']+(header['nrows']-0.5)*header['cellsize']
        x_all = np.linspace(x11, x11+(header['ncols']-1)*header['cellsize'],
                            header['ncols'])
        y_all = np.linspace(y11, y11-(header['nrows']-1)*header['cellsize'],
                            header['nrows'])
        row_all, col_all = sp.map2sub(x_all, y_all, self.header)
        rows, cols = np.meshgrid(row_all, col_all) # nrows*ncols array
        array = self.array[rows, cols]
        array = array.transpose()
        array = array.astype(self.array.dtype)
        obj_new = Raster(array=array, header=header)
        if hasattr(self, 'crs'):
            obj_new.crs = self.crs
        return obj_new
    
    def assign_to(self, new_header):
        """Assign_to the object to a new grid defined by new_header. 
        If their cellsizes are not equal, the original Raster will be 
        resampled to the target grid.

        Args:
            new_header (dict): Raster header dict

        Returns:
            Raster object: A newly defined grid
        """
        rows = np.arange(0, new_header['nrows'])
        cols = np.arange(0, new_header['ncols'])
        X, Y = sp.sub2map(rows, cols, new_header)
        grid_x, grid_y = np.meshgrid(X, Y)
        rows, cols = sp.map2sub(grid_x, grid_y, self.header)
        rows[rows > self.header['nrows']-1] = self.header['nrows']-1
        rows[rows < 0] = 0
        cols[cols > self.header['ncols']-1] = self.header['ncols']-1
        cols[cols < 0] = 0
        new_array = self.array[rows, cols]
        new_array = new_array+0.0
        new_array[new_array == self.header['NODATA_value']] = np.nan
        obj_new = Raster(array=new_array, header=new_header)
        if hasattr(self, 'crs'):
            obj_new.crs = self.crs
        return obj_new
    
    def paste_on(self, obj_large, ignore_nan=True):
        """Paste the object to a larger grid defined by obj_large and
        replace corresponding grid values with the object array, their 
        cellsizes MUST be equal

        Args:
            obj_large (Raster object): target object
            ignore_nan (bool, optional): indicate whether paste nan 
                values. Defaults to True.

        Returns:
            Raster object: target object with new array values pasted 
        """
        header_s = self.header
        header_l = obj_large.header
        x, y = self.to_points()
        r0, c0 = sp.map2sub(self.extent[0]+self.cellsize/2, 
                            self.extent[-1]-self.cellsize/2,
                            header_l)
        rows = np.arange(r0, r0+header_s['nrows'])
        cols = np.arange(c0, c0+header_s['ncols'])
        # cut array outside the range of the large raster grid
        ind_r = np.logical_and(rows > 0, rows <= header_l['nrows']-1)
        rows = rows[ind_r]
        ind_c = np.logical_and(cols > 0, cols <= header_l['ncols']-1)
        cols = cols[ind_c]
        array_small = self.array[ind_r, :]
        array_small = array_small[:, ind_c]
        rows_grid, cols_grid = np.meshgrid(rows, cols, indexing='ij')
        if ignore_nan:
            array_large = obj_large.array[rows_grid, cols_grid]
            ind_nan = np.isnan(array_small)
            array_small[ind_nan] = array_large[ind_nan]
        obj_large.array[rows_grid, cols_grid] = array_small
        return obj_large

    def to_points(self):
        """ Get X and Y coordinates of all raster cells

        Return:
            numpy array:  coordinates of the raster object
        """
        ny, nx = self.array.shape
        cellsize = self.header['cellsize']
        # coordinate of the centre on the top-left pixel
        x00centre = self.extent_dict['left'] + cellsize/2
        y00centre = self.extent_dict['top'] - cellsize/2
        x = x00centre+np.arange(nx)*cellsize
        y = y00centre-np.arange(ny)*cellsize
        xv, yv = np.meshgrid(x, y)
        return xv, yv
    
    def write_asc(self, output_file, compression=False, export_prj=False):
        """write raster as asc format file

        Args:
            output_file (string): output file path
            compression (bool, optional): _description_. Defaults to False.
            export_prj (bool, optional): indicate whether compress write the 
                asc file as gz. Defaults to False.
        """
        sp.arcgridwrite(output_file, self.array, self.header, compression)
        # if projection is defined, write .prj file for asc file
        if export_prj & hasattr(self, 'crs'):
            prj_file = output_file[0:-4]+'.prj'
            wkt = self.crs.to_wkt()
            with open(prj_file, "w") as prj:        
                prj.write(wkt)
    
    def write(self, output_file, compression=False):
        """Export to a file, tif, asc, txt, or gz

        Args:
            output_file (string): file name, ends with tif, asc, txt, or gz.
            compression (bool, optional): flag to indicate whether compress or not. 
                The default is False.. Defaults to False.
        """

        if output_file.endswith('.gz'):
            self.write_asc(output_file, compression=True)
        elif output_file.endswith('.tif'):
            self.write_tif(output_file)
        else:
            self.write_asc(output_file, compression)        
    
    def write_tif(self, output_file, src_epsg=27700, dtype='float64'):
        """ Convert to a rasterio dataset
        
        Args:
            output_file: a string to give output file name
            src_epsg: int scalar to give EPSG code of the coordinate reference
                system of the original dataset, default is 27700 for BNG

        """
        
        if output_file.endswith('.tif'):
            filename = output_file
        else:
            filename = output_file+'.tif'
        if 'int' in dtype:
            new_obj = self.to_int(dtype)
            array_data = new_obj.array+0
            new_obj.set_meta(src_epsg)
            meta = new_obj.meta
        else:
            array_data = self.array+0.0
            if not hasattr(self, 'meta'):
                self.set_meta(src_epsg)
            meta = self.meta # dictionary
        meta['dtype'] = array_data.dtype.name
        nomask = np.isnan(array_data)
        array_data[nomask] = meta['nodata']
        with rio.open(filename, 'w', **meta) as out_f:
            out_f.write(array_data, 1)
        meta['dtype'] = self.array.dtype.name
    
    def to_rasterio_ds(self):
        """
        convert object to a rasterio dataset

        Returns
        -------
        ds_rio : rasterio dataset
        """
        if not hasattr(self, 'meta'):
            self.set_meta()
        meta = self.meta
        filename = 'temp.tif'
        ds_rio = rio.open(filename, 'w+', **meta)
        ds_rio.write(self.array, 1)
        if os.path.isfile(filename):
            try:
                os.remove(filename)
            except:
                pass
        return ds_rio
    
    def set_meta(self, src_epsg=27700):
        """set rasterio meta data

        Args:
            src_epsg (int, optional): epsg code. Defaults to None.
        """
        from rasterio.transform import Affine
        dx = self.cellsize
        x = self.extent_dict['left'] # upper-left corner of the first pixel
        y = self.extent_dict['top']
        transform = Affine.translation(x, y)*Affine.scale(dx, -dx)
        if hasattr(self, 'crs'):
            crs = self.crs
        else:
            crs = rio.crs.CRS.from_epsg(src_epsg)
        ras_meta = {'driver': 'GTiff',
                    'dtype': self.array.dtype.name,
                    'nodata': self.header['NODATA_value'],
                    'width': self.array.shape[1],
                    'height': self.array.shape[0],
                    'count': 1,
                    'crs': crs,
                    'transform': transform}
        self.meta = ras_meta
        
    def set_nodata(self, NODATA_value):
        """Set nodata value of the object

        Args:
            NODATA_value (int): nodata value
        """
        self.nodata = NODATA_value
        self.header['NODATA_value'] = NODATA_value
        if hasattr(self, 'meta'):
            self.meta['nodata'] = NODATA_value
        
        
    def reproject(self, dst_epsg, output_file=None):
        """Reproject the raster to a different coordinate referenece system
        
        Args:
            output_file: a string to give output file name
            src_epsg: int scalar to give EPSG code of the coordinate reference
                system of the original dataset, default is 27700 for BNG
        Return:
            dst_rio: destination rasterio dataset
        
        """
        import rasterio
        from rasterio.warp import calculate_default_transform as cal_tsf
        from rasterio.warp import reproject, Resampling
        src_rio = self.to_rasterio_ds()
        dst_crs = rasterio.crs.CRS.from_epsg(dst_epsg)
        transform, width, height = cal_tsf(src_rio.crs, dst_crs, src_rio.width, 
                                           src_rio.height, *src_rio.bounds)
        kwargs = src_rio.meta.copy()
        kwargs.update({'crs': dst_crs, 'transform': transform, 'width': width,
                       'height': height, 'nodata':self.header['NODATA_value']})
        if output_file is None:
            filename = 'temp.tif'
        else:
            if output_file.endswith('.tif'):
                filename = output_file
            else:
                filename = output_file+'.tif'
        dst_rio = rasterio.open(filename, 'w+', **kwargs)
        reproject(source=rasterio.band(src_rio, 1),
                  destination=rasterio.band(dst_rio, 1),
                  src_transform=src_rio.transform,
                  src_crs=src_rio.crs,
                  dst_transform=transform,
                  dst_crs=dst_crs,
                  resampling=Resampling.nearest)
        if output_file is not None:
            dst_rio.close()
        return dst_rio

    def vectorize(self, valid_mask=None):
        """vectorize raster data array to a geopandas polygon dataframe which 
        can be written as ESRI shapefile or other formats

        Parameters
        ----------
        valid_mask : logical numpy array, optional
            define the cells to be rasterized. The default is None.

        Returns
        -------
        gdf : geopandas dataframe

        """
        transform = self.to_rasterio_ds().transform
        data_array = self.array.copy()
        
        if np.issubdtype(data_array.dtype, int):
            data_array = data_array.astype('int32')
        else: 
            data_array = data_array.astype('float32') 
        if valid_mask is None:
            mask = np.logical_and(~np.isnan(data_array),
                                  data_array != self.NODATA_value)
        else:
            mask = valid_mask
            
        output_polygons = rio_shapes(data_array, mask=mask, transform=transform) #
        polygon_collection = []
        value_collection = []
        for poly, val in output_polygons:
            oneshape = shapely_shape(poly)
            polygon_collection.append(oneshape.normalize())
            value_collection.append(val)
        features = [i for i in range(len(polygon_collection))]
        if hasattr(self, 'crs'):
            crs = self.crs
        else:
            crs = 'EPSG:27700'
        # add crs using wkt or EPSG to have a .prj file
        gdf = gpd.GeoDataFrame({'feature': features, 'geometry': polygon_collection}, 
                               crs=crs)
        gdf['value'] = value_collection
        return gdf
    
#%%=============================Visualization==================================
    def mapshow(self, **kwargs):
        """Display raster data without projection

        Args:
            figname: str, the file name to export map
            figsize: tuple, the size of map
            dpi: scalar, The resolution in dots per inch
            cax_str: str, the title of the colorbar
            relocate: True|False, relocate the origin of the grid to (0, 0)
            scale_ratio: 1|1000, axis unit 1 m or 1000 meter
            vmin: define the data range that the colormap covers
            vmax: define the data range that the colormap covers
            ytick_labelrotation: degree to rotate tick labels on y axis
            
        Example:
             mapshow(ax=ax, figname='my_fig', figsize=(6, 8), dpi=300,
                     title='My map', cax=True, cax_str='Meter', 
                     relocate=False, scale_ratio=1000, 
                     ytick_labelrotation=90)
        """
        fig, ax = gs.mapshow(raster_obj=self, **kwargs)
        return fig, ax
    
    def rankshow(self, **kwargs):
        """ Display array values in ranks
        
        Args:
            breaks: list of values to define rank. Array values lower than the
                first break value are set as nodata.
            color: color series of the ranks
            ylabelrotation: scalar giving degree to rotate yticklabel
            legend_kw: dict, keyword arguments to set legend. A colobar 
                     rather than a legend be displayed  if legend_kw is None.
            **kwargs: keywords argument of function imshow
        
        Example:
            rankshow(figname=None, figsize=None, 
                     dpi=200, ax=None, color='Blues', 
                     breaks=[0.2, 0.3, 0.5, 1, 2],
                     legend_kw={'loc':'upper left', 'facecolor':None, 
                                'fontsize':'small', 'title':'depth(m)', 
                                'labelspacing':0.1}, 
                     ytick_labelrotation=None, relocate=False, scale_ratio=1,
                     alpha=1)
        
        """
        fig, ax = gs.rankshow(self, **kwargs)
        return fig, ax
    
    def hillshade(self, **kwargs):
        """ Draw a hillshade map
        """
        fig, ax = gs.hillshade(self, **kwargs)
        return fig, ax

    def vectorshow(self, obj_y, **kwargs):
        """
        plot velocity map of U and V, whose values stored in two raster
        objects seperately
        """
        fig, ax = gs.vectorshow(self, obj_y, **kwargs)
        return fig, ax
#%% statistics function without nan
    def max(self, axis=None):
        # max value of the array ignoring nan values
        return np.nanmax(self.array, axis=axis)
    
    def min(self, axis=None):
        # min value of the array ignoring nan values
        return np.nanmin(self.array, axis=axis)
    
    def median(self, axis=None):
        # median value of the array ignoring nan values
        return np.nanmedian(self.array, axis=axis)
    
    def duplicate(self):
        """ duplicate the Raster object and return a new one so that change 
        the new object will not affect the original one
        """
        obj_duplicate = copy.deepcopy(self)
        return obj_duplicate

def from_tif(filename=None, array=None, meta=None):
    if filename is not None:
        obj_ras = Raster(filename)
    else:
        header = sp.meta2header(meta)
        obj_ras = Raster(array=array, header=header, crs=meta['crs'])
        obj_ras.meta = meta
    return obj_ras
    
#%%
def merge(obj_origin, obj_target, resample_method='bilinear'):
    """Merge the obj_origin to obj_target

    assign grid values in the origin Raster to the cooresponding grid cells in
    the target object. If cellsize are not equal, the origin Raster will be
    firstly resampled to the target object.
    
    Args:
        obj_origin: (Raster) original raster
        obj_target: (Raster) target raster
    """
    if obj_origin.header['cellsize'] != obj_target.header['cellsize']:
        obj_origin = obj_origin.resample(obj_target.header['cellsize'], 
                                   method=resample_method)
    grid_x, grid_y = obj_origin.to_points()
    rows, cols = sp.map2sub(grid_x, grid_y, obj_target.header)
    ind_r = np.logical_and(rows >= 0, rows <= obj_target.header['nrows']-1)
    ind_c = np.logical_and(cols >= 0, cols <= obj_target.header['ncols']-1)
    ind = np.logical_and(ind_r, ind_c)
    ind = np.logical_and(ind, ~np.isnan(obj_origin.array))
    obj_output = copy.deepcopy(obj_target)
    obj_output.array[rows[ind], cols[ind]] = obj_origin.array[ind]
    return obj_output

def main():
    """Main function
    """
    print('Class to deal with raster data')

if __name__=='__main__':
    main()
    
    

